/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include "client.h"

CLIENT_UL	*own;
dFloat		frametime;
TRACK_UL	*ptrack;		/* track to be played */
SIMPLECLIENT_UL	*firstcli;
Bool		paused;
UPDATE_UL	lastupdate;
#ifdef CLIENT
FILE		*sound;
CAMERA_UL	camera;
#endif

static void displayHelp ( void ) {
	fprintf( stderr,  
		"\nUsage: ulclient [options]\n"
		"Options are:\n"
		"* -client clientname\tno extension\n"
		"* -dname name\t\tdriver's name, should be AFTER -client\n"
#ifdef CLIENT
		"  -geometry widthxheight\n"
		"  -depth bits\t\t32, 24, 16, 8, or 0 for best\n"
		"  -single\t\tuse single buffering\n"
		"  -nodb\t\t\tdo not use direct buffers\n"
#endif
		"  -server address\n"
		"  -port number\n"
		"  -help\t\t\tdisplays this help\n"
		"Options marked with * must be present\n" );
}

static sInt args ( sInt argc, char **argv ) {
	sInt i;

	if ( argc == 1 ) {
		displayHelp();
		return FATALERROR;
	}
	for ( i = 1; i < argc; i++ ) {
		if ( strcmp ( argv[i], "-server" ) == 0 )
			strcpy( prefs->servername, argv[++i] );
#ifdef CLIENT
		else if ( strcmp ( argv[i], "-geometry" ) == 0 )
			sscanf( argv[++i], "%dx%d", prefs->width, prefs->height );
		else if ( strcmp ( argv[i], "-depth" ) == 0 ) {
			prefs->depth = atoi( argv[++i] );
			if ( prefs->depth != 32 && prefs->depth != 24 &&
			     prefs->depth != 16 && prefs->depth != 8 &&
			     prefs->depth != 0 ) {
				fprintf( stderr, "Invalid depth mode: %d",
						prefs->depth );
				goto argsError;
			}
		}
		else if ( strcmp ( argv[i], "-single" ) == 0 )
			prefs->single = True;
		else if ( strcmp ( argv[i], "-nodb" ) == 0 )
			prefs->usedb = False;
#endif
		else if ( strcmp ( argv[i], "-port" ) == 0 )
			prefs->serverport = atoi( argv[++i] );
		else if ( strcmp ( argv[i], "-client" ) == 0 ) {
			if ( (own = readClient( argv[++i] )) == NULL )
				goto argsError;
		}
		else if ( strcmp ( argv[i], "-dname" ) == 0 ) {
			if ( !own )
				goto argsError;
			own->driver = strdup( argv[++i] );
		}
		else if ( strcmp ( argv[i], "-help" ) == 0 ) {
			displayHelp();
			return FATALERROR;
		}
		else {
argsError:
			fprintf( stderr, "Error at argument %d: %s\n", 
				i, argv[i] );
			displayHelp();
			return FATALERROR;
		}
	}
	if ( own == NULL ) {
		fprintf( stderr, "You need to suply client plugin name!\n" ); 
		return FATALERROR;
	}
	if ( own->driver == NULL ) {
		fprintf( stderr, "You need to suply client's \"driver\" name!" );
		return FATALERROR;
	}
	return OK;
}

static SIMPLECLIENT_UL *createSimpleClient ( void ) {
	SIMPLECLIENT_UL *cli, *tmp;
	if ( !(cli = (SIMPLECLIENT_UL *)malloc(sizeof(SIMPLECLIENT_UL))) )
		return NULL;
	cli->next = NULL;
	if ( firstcli == NULL ) {
		firstcli = cli;
		return cli;
	}
	for ( tmp = firstcli; tmp->next != NULL; tmp++ ) ;
	tmp->next = cli;
	return cli;
}

static sInt initnet ( void ) {
	CLIENTM_UL		req;
	SERVERM_UL		*resp;
	sInt 			i, n;
	uByte			buffer[512], *p[5];
	FILE 			*dl;
	SIMPLECLIENT_UL		*cli;

	/* connect to the xraced server */
	if ( netclient( prefs->servername ) == FATALERROR ) {
		fprintf(stderr, "Could not connect to %s\n", prefs->servername );
		exit( 2 );
	}
	
	netaddsock( server );

	req.request = UL_HELLO;
#ifdef CLIENT
	/* this is a quick fix. In future, AI may be supported in CLIENTs too.*/
	req.hello.ai = False;
#else
	/* NONE are always AI */
	req.hello.ai = True;
#endif
	req.hello.driver = own->driver;
	req.hello.model = own->model;
	copyColor( &req.hello.color, &own->color );
	netclientsend( &req );
	netflush();

	/* wait for the "welcome" message. */
	do {
		resp = netclientrcv();
		if ( !resp ) {
			fprintf( stderr,
			   "Error waiting for welcome message from server\n" );
			return FATALERROR;
		}
	} while ( resp->response != UL_WELCOME );

	own->id = (((uLong)resp->welcome.number)<<16)&0xFFFF0000;

	/* is there a track? */
	switch ( findTrack( resp->welcome.track ) ) {
		case FATALERROR:
			return FATALERROR;
		case ERROR:
/************* should be reviewed */
			/* we don't have this track */
			if ( prefs->dltrack == True ) {
				req.request = UL_GETFILE;
				req.getfile.name = buffer;
				strcpy( buffer, TRACKDIR );
				strcat( buffer, resp->welcome.track );
				netclientsend( &req );
				netflush();
				/* receive track */
				do {
					resp = netclientrcv();
					if ( !resp ) {
						fprintf( stderr, 
						   "Error waiting for welcome message from server\n" );
						return FATALERROR;
					}
				} while (resp->response != UL_FILE );
				sprintf( buffer, "%s%s%s", ULDIR, TRACKDIR,
						resp->welcome.track );
				dl = fopen( buffer, "w" );
				if ( !dl ) {
					fprintf( stderr, "Cannot save track in %s",
							buffer );
					return FATALERROR;
				}
				if ( fwrite( resp->file.data, resp->file.size, 1, dl )
						!= resp->file.size ) {
					fprintf( stderr, "Error saving track in %s",
							buffer );
					exit( 2 );
				}
				fclose( dl );
				goto readtrack;
			}
			else {
				fprintf( stderr, "Track %s not found.\n",
						resp->welcome.track );
				return FATALERROR;
			}
		case OK:
readtrack:
			if ( (ptrack = readTrack( resp->welcome.track ))
					== NULL ) 
				exit( 2 );
	}

	copyV( &own->position, &resp->welcome.position );
	copyV( &own->direction, &resp->welcome.direction );
	own->angle = resp->welcome.angle;
}

#ifdef CLIENT
static sInt handleggiEvent ( void ) {
	uint32			key;
	uByte			buffer[20];
	sInt			count;
        ggi_event		ev;
        ggi_event_mask		mask;
	struct timeval		t;

	while (1) {
		mask = ggiEventPoll( vis, emAll, &t );
		if (!mask) return;
		ggiEventRead( vis, &ev, emAll );
		switch ( ev.any.type ) {
		  case evExpose:
			render( &camera );
		  	break;

		  case evKeyPress:
			own->event( KEYPRESS_UL, ev.key.label, ev.key.sym );
			break;

		  case evKeyRepeat:
			own->event( KEYREPEAT_UL, ev.key.label, ev.key.sym );
			break;
#ifdef OLD
/***** this will be dealt by the client plugin, but is resting here by now.
Also, hardcoding the keys is horrible */
			switch ( ev.key.label ) {
			  case GII_UCEscape: /* open menu? or quit? */
				break;
			  case 'p': /* pause */
			  case 'P':
				paused = ( paused == True ? False : True );
				break;
			  case 'q': /* quit */
			  case 'Q':
				/* do you really want to quit etc */
				return FATALERROR;
				break;
			  case 's': /* sound on/off */
			  case 'S':
				prefs.sound = ( prefs.sound == True ? False : True );
				break;
			  case 'm': /* map on/off */
			  case 'M':
				break;
			}
#endif
			break;

		  case evKeyRelease:
			own->event( KEYRELEASE_UL, ev.key.label, ev.key.sym );
			break;
		}
	}
}
#endif

int main ( int argc, char **argv ) {
	sInt 				sock;
	struct timeval			t1, t2;
	SERVERM_UL 			*resp;
	CLIENTM_UL			req;
	TIMESTAMP_UL			ts;

#ifdef CLIENT
	fprintf( stdout, "UNLIMITED SIMULATOR Graphic Client, Version %s (%s).\n",
#else
	fprintf( stdout, "UNLIMITED SIMULATOR Simple Client, Version %s (%s).\n",
#endif
			VERSION, BUILDDATE );

	loadprefs();
	fprintf( stdout, "Checking args... " );
	if ( args( argc, argv ) == FATALERROR )
		exit(1);
	fprintf( stdout, "OK\n" );

#ifdef CLIENT
	fprintf( stdout, "Starting engine... " );
	if ( initULEngine( prefs->width, prefs->height, prefs->depth, prefs->single,
			prefs->usedb ) == FATALERROR ) {
		endULEngine(); /* this will finish ggi */
		exit(2);
	}
	fprintf( stdout, "OK\n" );

	if ( prefs->sound == True ) {
		fprintf( stdout, "Starting sound... " );
		if ( !(sound = popen( "unixsnd", "w" ) ) ) {
			fprintf( stdout, "FAILED\n" );
			fprintf( stdout, "Couldn't open sound... We're mute.\n" );
			prefs->sound = False;
		}
		else
			fprintf( stdout, "OK\n" );
	}
#endif

	fprintf( stdout, "Starting net... " );
	if ( initnet() == FATALERROR ) {
		fprintf( stdout, "FAILED\n" );
		goto TheEnd;
	}
	fprintf( stdout, "OK\n" );

#ifdef CLIENT
	fprintf( stdout, "Starting font server... " );
	initFont();
	fprintf( stdout, "OK\n" );
#endif
	/***** show open screen and menu */
	
	lastupdate.time = timestamp( NULL );
	copyV( &lastupdate.lastpos, &own->position );
	copyV( &lastupdate.lastdir, &own->direction );
	lastupdate.lastangle = own->angle;
	
	/* and the race begins */
	fprintf( stdout, "Starting simulation... " );
	do {
		gettimeofday( &t1, NULL );

#ifdef HAVETOCHANGE
		sock = netwait( servermode == NRMODE_RACE ? -1L : 0L );
#endif

		if ( sock == server ) {
			/* message from server */
			do {
				resp = netclientrcv();
				if ( !resp )
					goto TheEnd;
				doresp(resp);
			} while ( netclientmore() );
		}
#ifdef CLIENT
		handleggiEvent();
#endif

		if ( paused == True )
			continue;

		/* cd ( firstdoodad, own, force?, torque? );  col det*/
		
#ifdef CLIENT
		setCameraV( &camera, &own->position, &own->direction, own->angle );
		render( &camera );
#endif
		
		gettimeofday( &t2, NULL );
		frametime = timeval_subtract( &t1, &t2, NULL );

		own->loop();
		
		/* check frame time and send updates to server */
		ts = timestamp(NULL);
		if ( ts - lastupdate.time > prefs->updatefreqts ) {
			/* send data */
			if ( lastupdate.n < prefs->maxupdates ) {
				req.request = UL_CONTROL;
				req.control.dx = (lastupdate.lastpos.x - own->position.x)*100;
				req.control.dy = (lastupdate.lastpos.y - own->position.y)*100;
				req.control.dz = (lastupdate.lastpos.z - own->position.z)*100;
				req.control.dwx = (lastupdate.lastdir.x - own->direction.x)*100;
				req.control.dwy = (lastupdate.lastdir.y - own->direction.y)*100;
				req.control.dwz = (lastupdate.lastdir.z - own->direction.z)*100;
				req.control.dangle = (lastupdate.lastangle - own->angle)*100;
				lastupdate.n++;
			}
			else {
				lastupdate.n = 0;
				req.request = UL_PRECISECONTROL;
				copyV( &req.precisecontrol.position, &own->position );
				copyV( &req.precisecontrol.direction, &own->direction );
				req.precisecontrol.angle = own->angle;
			}
			req.control.timestamp = (uShort)(ts&0xFFFF);
			netclientsend( &req );

			/* update lastupdate! */
			lastupdate.time = timestamp( NULL );
			copyV( &lastupdate.lastpos, &own->position );
			copyV( &lastupdate.lastdir, &own->direction );
			lastupdate.lastangle = own->angle;
		}
	} while ( 0 == 0 );

TheEnd:
	free( prefs );
#ifdef CLIENT
	if ( sound )
		pclose( sound );
	endULEngine();
#endif
	return 0;
}
