/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

/* This is a simple vehicle to walk around, with 6 DOF. It's provided as a
 * skeleton of a vehicle plugin: adapt it as you want and create your own
 * vehicles!
 */

#include "client.h"

/*
 * this function initializes all data. Here you read the vehicle's specs,
 * initialize variables, etc. This function is called once when the vehicle is
 * loaded. Do not rename it. the u_Vehicle strucutre is already allocated and
 * initialized.
 */
sInt u_initVehicle ( u_Vehicle *v ) {

	/* here you should malloc v->u_intData */

	return OK;
}

/* this function frees all data. It is called once when the simulation is
 * over, to clean up everything. Do not rename it. 
 */
sInt u_endVehicle ( u_Vehicle *v ) {

	/* here you should free v->u_intData */

	return OK;
}

/* When some event is received, such as a key press, mouse click, etc, which is
 * relevant to the client, it's passed to this function. 
 */

#define SPEED 		8	/* m/s */
#define ANGSPEED	PI/18	/* rad/s */

sInt u_handleEvent ( u_Vehicle *v, sInt key ) {
	char pos[100];
	switch ( event ) {
#ifndef AI_ONLY
	  case KEYPRESS_UL:
	  case KEYREPEAT_UL:
		switch ( label ) {
			/* these are temporary; we are using world coordinates
			instead of local. */
			/* translation */
		        case GIIK_Up:
				own->position.x += SPEED/frametime;
				break;
			case GIIK_Down:
				own->position.x -= SPEED/frametime;
				break;
		        case GIIK_Left:
				own->position.y += SPEED/frametime;
				break;
		        case GIIK_Right:
				own->position.y -= SPEED/frametime;
				break;
			case 'a':
			case 'A':
				own->position.z += SPEED/frametime;
				break;
			case 'z':
			case 'Z':
				own->position.z -= SPEED/frametime;
				break;

			/* rotation */
		        case GIIK_Insert:
				own->direction.x += ANGSPEED/frametime;
				break;
			case GIIK_Delete:
				own->direction.x += ANGSPEED/frametime;
				break;
		        case GIIK_Home:
				own->direction.y += ANGSPEED/frametime;
				break;
		        case GIIK_End:
				own->direction.y += ANGSPEED/frametime;
				break;
			case GIIK_PageUp:
				own->direction.z += ANGSPEED/frametime;
				break;
			case GIIK_PageDown:
				own->direction.z += ANGSPEED/frametime;
				break;
			
	        	case GIIK_Enter:
				sprintf( pos, "(%f,%f,%f)", own->position.x,
					own->position.y, own->position.z );
				addmessage( pos, -1 );
				break;
				
			case GIIUC_Escape:
				return FATALERROR;

			default:
				break;
		}
		break;
	
	  case KEYRELEASE_UL:
		break;
#endif /* AI_ONLY */
	  /* other events in future */
	}
	return OK;
}


/*
 * this is the main loop. It's called every loop of the client main program,
 * and here you should do: physics simulation, AI, etc. Note that collision 
 * detection is performed automatically by the world plugin, which passes a 
 * event to the u_handleEvent() function. Note that, although mainLoop() and
 * handleEvent() are different functions, you can place events in a stack and
 * handle they here.
 */

sInt mainLoop ( u_Vehicle *v ) {
	/* nothing here! :*/
	return OK;
}
