/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#ifndef U_GEOMETRY_H
#define U_GEOMETRY_H

/* have to fix the u_gType [] * problem. Is u_gType really needed? */


/********************************
 ** general definition (parent) *
 ********************************/

typedef const char * u_gType;

struct u_geometry {
	u_gType			type;

	u_Vector		position, direction;
	ffloat			angle;
	u_Sphere		bouding;
	uByte			*name;
};
typedef struct u_geometry u_Geometry;

/* 
 * functions in geometry.c
 */

u_Geometry *		u_createGeometry
		( char *name );
void			u_freeGeometry
		( u_Geometry *geo );


/***********************
 ** wireframe geometry *
 ***********************/

extern u_gType u_gType_Wireframe;

struct point_ul {
	u_Vector		*vect;
	u_Vector		*norm;		/* normal, if point belongs to a 
					   surface. */
#ifdef CLIENT
	fFloat			u, v;		/* texture mapping coordinates */
#endif

};
typedef struct u_point u_Point;


struct u_polygon {
	uByte			npoints;  	/* no of point in polygon */
	u_Point 		**points;
	u_Vector	 	normal;
	u_Vector		position;
	
#ifdef OLD_2B_UPDATED_IN_FUTURE_AND_CLIENT
	ggi_color 	*color;		/* one for each vertex */
	ggi_color	*flatcolor;	/* if FLAT is used, this is the 
					   average of all vertices' colors
					   + light info */
	Bool		onecolor;	/* if every point has the same color. */
	IMAGE_UL	*texture;
#endif

	Bool			visible;
};
typedef struct u_polygon u_Polygon;

struct u_wireframe {
	/* this line must be the first in structure */
	u_Geometry		parent;
	
	u_gType			gtype;
	
	u_Polygon		**polygons;
	Int			npoly, nvert;
	Int			ncopies;
};
typedef struct u_wireframe u_Wireframe;

/* 
 * functions in wireframe.c
 */

u_Wireframe *		u_createWireframe 
		( char *name );
void			u_freeWireframe 
		( u_Wireframe *wf );
void			u_freePolygon 
		( u_Polygon *p );

/************************
 ** volumetric geometry *
 ************************/

extern u_gType u_gType_Volumetric;

struct u_voxel {
	ffloat			opacity;
	/* in future: color, etc */
};
typedef struct u_voxel u_Voxel;

struct u_volumetric {
	/* this line must be the first in structure */
	u_Geometry		parent;
	
	sInt			x, y, z;	/* number of voxels */
	u_Voxel			***voxel;
};
typedef struct u_volumetric u_Volumetric;

/* 
 * functions in volumetric.c
 */

/*******************
 ** NURBS geometry *
 *******************/

extern u_gType u_gType_NURBS[];

struct u_nurbs {
};
typedef struct u_nurbs u_Nurbs;

/* 
 * functions in volumetric.c
 */

#endif
