/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#include <dlfcn.h>
#include <stdio.h>
#include "unreality.h"

u_Net  			*netplugin;
u_Renderer		*rplugin;
u_SoundEngine		*splugin;
u_World			*wplugin;

/* loads a plugin named name, of type type, in structure, which is hopefully of
the correct kind (no checking made yet). Calls the plugin initializer,
dlopen will call _init.
Global plugins are freed and malloced. Objects and vehicles are not. */

sInt u_loadPlugin ( char *name, u_Plugin type, void *structure ) {
	char 			*error, pathname[512];
	u_Vehicle		*vplugin;
	u_Object		*oplugin;
	
	switch ( type ) {
	   case P_NET:
		if ( netplugin )
			u_closePlugin( netplugin, NET );
		netplugin = (u_Net *)malloc(sizeof(u_Net));

		sprintf( pathname, "%s/plugins/network/%s.so", DATA_PATH, name );

		netplugin->u_lib = dlopen( pathname, RTLD_LAZY | RTLD_GLOBAL );
		if ( netplugin->u_lib == NULL ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		
		netplugin->u_initNet = dlsym( netplugin->u_lib, "u_initNet");
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		netplugin->u_endNet = dlsym( netplugin->u_lib, "u_endNet");
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		netplugin->u_readNet = dlsym( netplugin->u_lib, "u_readNet");
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		netplugin->u_writeNet = dlsym( netplugin->u_lib, "u_writeNet");
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}

		netplugin->u_initNet( netplugin );
		structure = (void *)netplugin;
		break;

	   case P_RENDERER:
		if ( rplugin )
			u_closePlugin( rplugin, RENDERER );
		rplugin = (u_Renderer *)malloc(sizeof(u_Renderer));

		sprintf( pathname, "%s/plugins/renderer/%s.so", DATA_PATH, name );
		
		rplugin->u_lib = dlopen(pathname, RTLD_LAZY | RTLD_GLOBAL);
		if ( rplugin->u_lib == NULL ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		
		rplugin->u_initRenderer   = dlsym( rplugin->u_lib, "u_initRenderer" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_endRenderer	 = dlsym( rplugin->u_lib, "u_endRenderer" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_render	 = dlsym( rplugin->u_lib, "u_render" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_setCamera	 = dlsym( rplugin->u_lib, "u_setCamera" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_createGeometry = dlsym( rplugin->u_lib, "u_createGeometry" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_deleteGeometry = dlsym( rplugin->u_lib, "u_deleteGeometry" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_createLight	 = dlsym( rplugin->u_lib, "u_createLight" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		rplugin->u_deleteLight	 = dlsym( rplugin->u_lib, "u_deleteLight" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		vplugin->u_printf = dlsym( rplugin->u_lib, "u_printf" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}

		rplugin->u_initRenderer( rplugin );
		structure = (void *)rplugin;
		break;

	   case P_SOUNDENGINE:
		if ( splugin )
			u_closePlugin( splugin, SOUND );
		splugin = (u_SoundEngine *)malloc(sizeof(u_SoundEngine));

		sprintf( pathname, "%s/plugins/sound/%s.so", DATA_PATH, name );

		splugin->u_lib = dlopen( pathname, RTLD_LAZY | RTLD_GLOBAL );
		if ( splugin->u_lib == NULL ) {
			fprintf( stderr, "%s", dlerror() );
			return;
		}
		splugin->u_initSound   = dlsym( splugin->u_lib, "u_initSound" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror() );
			return FATALERROR;
		}
		splugin->u_endSound    = dlsym( splugin->u_lib, "u_endSound" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror() );
			return FATALERROR;
		}
		splugin->u_playSound   = dlsym( splugin->u_lib, "u_playSound" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror() );
			return FATALERROR;
		}
		splugin->u_stopSound   = dlsym( splugin->u_lib, "u_stopSound" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror() );
			return FATALERROR;
		}
		splugin->u_pauseSound  = dlsym( splugin->u_lib, "u_pauseSound" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror() );
			return FATALERROR;
		}
		splugin->u_resumeSound = dlsym( splugin->u_lib, "u_resumeSound" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror() );
			return FATALERROR;
		}

		splugin->u_initSound( splugin );
		structure = (void *)splugin;

		break;
	   case P_WORLD:
		if ( wplugin )
			u_closePlugin( wplugin, WORLD );
		wplugin = (u_World *)malloc(sizeof(u_World));

		sprintf( pathname, "/%s/plugins/world/%s.so", DATA_PATH, name );

		wplugin->u_lib = dlopen(pathname, RTLD_LAZY | RTLD_GLOBAL);
		if ( wplugin->u_lib == NULL ) {
			fprintf( stderr, "%s", dlerror());
			return OK;
		}
		
		wplugin->u_initWorld = dlsym( wplugin->u_lib, "u_initWorld" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		wplugin->u_endWorld  = dlsym( wplugin->u_lib, "u_endWorld" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		wplugin->u_loop	    = dlsym( wplugin->u_lib, "u_loop" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}

		wplugin->u_initWorld( wplugin );		
		structure = (void *)wplugin;

		break;
	   case P_VEHICLE:
		if ( !structure ) {
			fprintf( stderr, "NULL vehicle structure in loadplugin\n" );
			return ERROR;
		}
		vplugin = (u_Vehicle *)structure;

		sprintf( pathname, "%s/plugins/vehicle/%s.so", DATA_PATH, name );

		vplugin->u_lib = dlopen(name, RTLD_LAZY | RTLD_GLOBAL);
		if ( vplugin->u_lib == NULL ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}

		vplugin->u_initVehicle  = dlsym( vplugin->u_lib, "u_initVehicle" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		vplugin->u_endVehicle   = dlsym( vplugin->u_lib, "u_endVehicle" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		vplugin->u_handleEvent  = dlsym( vplugin->u_lib, "u_handleEvent" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		vplugin->u_loop	       = dlsym( vplugin->u_lib, "u_loop" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}

		vplugin->u_initVehicle(vplugin);
		structure = (void *)vplugin;
		
		break;
	   case P_OBJECT:
		if ( !structure ) {
			fprintf( stderr, "NULL vehicle structure in loadplugin\n" );
			return ERROR;
		}
		oplugin = (u_Object *)structure;

		sprintf( pathname, "%s/plugins/object/%s.so", DATA_PATH, name );

		oplugin->u_lib = dlopen(name, RTLD_LAZY | RTLD_GLOBAL);
		if ( oplugin->u_lib == NULL ) {
			/* no plugin for this object */
			oplugin->u_animated = False;
			return OK;
		}

		oplugin->u_initObject = dlsym( oplugin->u_lib, "u_initObject" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		oplugin->u_endObject  = dlsym( oplugin->u_lib, "u_endObject" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}
		oplugin->u_loop	     = dlsym( oplugin->u_lib, "u_loop" );
		if ( ((error = dlerror()) != NULL) ) {
			fprintf( stderr, "%s", dlerror());
			return FATALERROR;
		}

		oplugin->u_animated = True;
		oplugin->u_initVehicle(oplugin);
		structure = (void *)oplugin;

		break;
	}
	return OK;
}

/* Closes the lib of structure, which is hopefully of type type. Will call
the plugin end function, dlclose will call _fini.
If structure is global (as RENDERER), which is determined by the type argument,
the structure argument can be NULL. */

/* SHOULD WE close all VehExtras and ObjExtras? */
void u_closePlugin ( void *structure, u_Plugin type ) {
	u_Vehicle 		*vplugin;
	u_Object 		*oplugin;
	int i;

	switch ( type ) {
	   case P_NET:
		if ( !netplugin )
			return;
		netplugin->endNet();
		dlclose( netplugin->u_lib );
		free( netplugin );
		break;

	   case P_RENDERER:
		if ( !rplugin )
			return;
		rplugin->endRenderer();
		dlclose( rplugin->u_lib );
		free( rplugin );
		break;

	   case P_SOUNDENGINE:
		if ( !splugin )
			return;
		splugin->endSound();
		dlclose( splugin->u_lib );
		free( splugin );
		break;

	   case P_WORLD:
		if ( !wplugin )
			return;
		wplugin->endWorld();
		dlclose( wplugin->u_lib );
		free( wplugin );
		break;

	   case P_VEHICLE:
		if ( !vplugin )
			return;
		vplugin = (u_Vehicle *)structure;
		netplugin->endVehExtra( vplugin->extra[NET] );
		rplugin->endVehExtra( vplugin->extra[RENDERER] );
		splugin->endVehExtra( vplugin->extra[SOUNDENGINE] );
		wplugin->endVehExtra( vplugin->extra[WORLD] );
		dlclose( vplugin->u_lib );
		free( vplugin );
		break;

	   case P_OBJECT:
		if ( !oplugin )
			return;
		oplugin = (u_Object *)structure;
		netplugin->endObjExtra( oplugin->extra[NET] );
		rplugin->endObjExtra( oplugin->extra[RENDERER] );
		splugin->endObjExtra( oplugin->extra[SOUNDENGINE] );
		wplugin->endObjExtra( oplugin->extra[WORLD] );
		dlclose( oplugin->u_lib );
		free( oplugin );
		break;

	}
}
