/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#ifndef UNREALITY_H
#define UNREALITY_H

#include "config.h"
#include "events.h"
#include "geometry.h"

struct u_net {
	void 			*u_lib;

	sInt			(* u_initNet)
		( struct u_net * );
	char *			(* u_readNet)
		( void );

	sInt			(* u_startObjExtra)
		( void *extra );
	sInt			(* u_freeObjExtra)
		( void *extra );

	sInt			(* u_startVehExtra)
		( void *extra );
	sInt			(* u_freeVehExtra)
		( void *extra );

	sInt			(* u_writeNet)
		( char *data );
	sInt			(* u_endNet)
		( struct u_net * );
	
	struct u_intNet		*intData;
};
typedef struct u_net u_Net;
 
struct u_renderer {
	void			*u_lib;

	/* initialization stuff */
	sInt 			(* u_initRenderer)
		( struct u_renderer * );
	sInt			(* u_endRenderer)
		( struct u_renderer * );

	/* object and vehicle stuff */
	sInt			(* u_startObjExtra)
		( void *extra );
	sInt			(* u_freeObjExtra)
		( void *extra );

	sInt			(* u_startVehExtra)
		( void *extra );
	sInt			(* u_freeVehExtra)
		( void *extra );

	/* rendering */
	sInt			(* u_render)
		( void );
	void			(* u_setCamera)
		( u_Vector, u_Vector, float );

	/* geometry functions */
	uLong			(* u_addGeometry)
		( u_Geometry * );
	sInt			(* u_deleteGeometry)
		( uLong id );
	void			(* u_transformGeometry)
		( u_Transform *, uLong id );

	uLong			(* u_createLight)
		( u_Vector, u_Vector, float, float, float );
	sInt			(* u_deleteLight)
		( uLong  );

	sInt			(* u_printMessage)
		( char *message, ... );

	/* internal data */
	struct u_intRenderer	*intData;
};
typedef struct u_renderer u_Renderer;

struct u_soundEngine {
	void 			*u_lib;

	sInt			(* u_initSoundEngine)
		( struct u_soundEngine * );
	sInt			(* u_endSoundEngine)
		( struct u_soundEngine * );

	sInt			(* u_startObjExtra)
		( void *extra );
	sInt			(* u_freeObjExtra)
		( void *extra );

	sInt			(* u_startVehExtra)
		( void *extra );
	sInt			(* u_freeVehExtra)
		( void *extra );

	Sound *			(* u_playSound)
		( void *data );
	sInt			(* u_stopSound)
		( Sound *sound );
	sInt			(* u_pauseSound)
		( Sound *sound );
	sInt			(* u_resumeSound)
		( Sound *sound );

	struct u_intSoundEngine	*u_intData;
};
typedef struct u_soundEngine u_SoundEngine;

struct u_world {
	void 			*u_lib;

	sInt			(* u_initWorld)
		( struct u_world * );
	sInt			(* u_endWorld)
		( struct u_world * );

	sInt			(* u_startObjExtra)
		( void *extra );
	sInt			(* u_freeObjExtra)
		( void *extra );

	sInt			(* u_startVehExtra)
		( void *extra );
	sInt			(* u_freeVehExtra)
		( void *extra );

	sInt 			(* u_loop)
		( void *data ); /* to provide dynamic world */

	u_Geometry		*u_geom;

	struct u_intWorld	*u_intData;
};
typedef struct u_world u_World;

struct u_vehicle {
	void 			*u_lib;

	sInt			(* u_initVehicle)
		( struct u_vehicle * );
	sInt			(* u_endVehicle)
		( struct u_vehicle * );

	sInt			(* u_handleEvent)
		( struct u_vehicle *, u_Event *type, void *data );
	sInt			(* u_loop)
		( struct u_vehicle * );
	
	/* attributes */
	uShort			id;
	u_Geometry		*u_geom;
	bool			u_human;

	void 			*u_extra[4];
	struct u_intVehicle	*u_intData;
};
typedef struct u_vehicle u_Vehicle;

struct u_object {
	void		 	*u_lib;

	sInt			(* u_initObject)
		( struct u_object * );
	sInt			(* u_endObject)
		( struct u_object * );

	sInt			(* u_handleEvent)
		( struct u_object *, u_Event *type, void *data );
	sInt		 	(* u_loop)
		( struct u_object * );

	/* attributes */
	uLong			id;
	u_Geometry		*u_geom;
	bool			u_animated;

	void 			*u_extra[4];
	struct u_intObject	*u_intData;
};
typedef struct u_object u_Object;

struct u_sitem {
	void 		*data;
	struct u_sitem	*next;
};
typedef struct u_sitem u_sItem;

struct u_slist {
	u_Item 		*head;
	u_Item		*current;
	int		total;
};
typedef struct u_list u_sList;

extern	u_Net 			*netplugin;
extern	u_Renderer 		*rplugin;
extern	u_SoundEngine 		*splugin;
extern	u_World			*wplugin;

/* plugin.c */
sInt 		u_loadPlugin ( void *structure, char *name, u_Plugin type );
void		u_closePlugin ( void *structure, u_Plugin type ); 

/* object.c */
u_Object *	u_createObject 
			( char *name );
void 		u_freeObject
			( u_Object *obj );

/* vehicle.c */
u_Vehicle *	u_createVehicle
			( char *name );
void 		u_freeVehicle
			( u_Vehicle *veh );

/* list.c */
sint		u_addsItem
			( u_sList *list, void *data );
void *		u_removesItem
			( u_sList *list, u_sItem *item );
sint		u_removesItembyData
			( u_sList *list, void *data );
void *		u_headsData
			( u_sList *list );
void *		u_nextsData
			( u_sList *list );


#endif
