/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

/* analytic geometry stuff
*/
#include "mathlib.h"

void initPlaneP ( PLANE_UL *p, POINT_UL *p1, POINT_UL *p2, POINT_UL *p3 ) {
        VECTOR_UL v1, v2;
	v1.x = p2->vect->x - p1->vect->x;
	v2.x = p3->vect->x - p1->vect->x;
	v1.y = p2->vect->y - p1->vect->y;
	v2.y = p3->vect->y - p1->vect->y;
	v1.z = p2->vect->z - p1->vect->z;
	v2.z = p3->vect->z - p1->vect->z;
	crossprodV ( &v1, &v2, &(p->normal) );
	normalizeV( &(p->normal) );

	p->dist = dotprodV( &(p->normal), p1->vect );
	
	p->a = p->normal.x;
	p->b = p->normal.y;
	p->c = p->normal.z;
	p->d = -(p->a*p1->vect->x + p->b*p1->vect->y + p->c*p1->vect->z);
	
	copyV( &p->p, p1->vect );
}

void initPlaneVD ( PLANE_UL *p, VECTOR_UL *v, double d ) {
	copyV( &(p->normal), v );
	normalizeV( &(p->normal) );
	p->dist = d;
	
	p->a = p->normal.x;
	p->b = p->normal.y;
	p->c = p->normal.z;
	
	/* missing the rest */
}

/* given a plane an a straight line in format v1+h.v2, find the intersection.
Returns 0 if no intersection found, 1 otherwise. */
int intPlaneLine ( PLANE_UL *p, VECTOR_UL *v1, VECTOR_UL *v2, VECTOR_UL *r ) {
	double		h, t;

	if ( (t = p->a*v2->x + p->b*v2->y + p->c*v2->z) == 0 ) /* determinant vanishes */
		return 0;

	h = -(p->d + p->a*v1->x + p->b*v1->y + p->c*v1->z)/t;
	r->x = v1->x + h*v2->x;
	r->y = v1->y + h*v2->y;
	r->z = v1->z + h*v2->z;
	return 1;
}
