/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#include "mathlib.h"

/* swaps line A with line B */
static void changeline ( int n, int m, double matrix[n][m], int a, int b ) {
	int i;
	double d;
	for ( i = 0; i < n; i++ ) {
		d = matrix[a][i];
		matrix[a][i] = matrix[b][i];
		matrix[b][i] = d;
	}
}

/* solve linear system using Gauss elimination */
int gausselim ( int n, double A[n][n], double B[n], double x[n] ) {
	int i, j, k;
	double factor;

	for ( j = 0; j < n; j++ ) { // columns
		for ( i = n-1; i > j; i-- ) {
			factor = A[i][j]/A[j][j];
			for ( k = 0; k < n; k++ )
			    A[i][k] -= factor*A[j][k];
			B[i] -= factor*B[j];
		}
		/* is system ok? */
		if ( A[j+1][j+1] == 0.0 && j+1 < n) {
			for ( k = j+2; k < n; k++ )
				if ( A[k][j+1] != 0 )
					break;
			if ( k == n || j+2 == n ) /* undetermined */
				return -1;
			changeline( n, n, A, j+1, k );
			changeline( n, 1, B, j+1, k );
		}
	}

	/* solve system */
	for ( i = n-1; i >= 0; i-- ) {
		x[i] = B[i];
		for ( j = i+1; j < n; j++ )
			x[i] -= (A[i][j]*x[j]);
		x[i] /= A[i][i];
	}
	return 0;
}
