/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#include "mathlib.h"

/* assumes that quaternions are allocated */

void initQuatD ( u_Quaternion *q, fFloat s, fFloat x, fFloat y, fFloat z ) {
	q->s = s;
	q->v.x = x;
	q->v.y = y;
	q->v.z = z;
}

void initQuatVS ( u_Quaternion *q, fFloat s, u_Vector *v ) {
	q->s = s;
	q->v.x = v->x;
	q->v.y = v->y;
	q->v.z = v->z;
}

/* returns normalized */
void initQuatVA ( u_Quaternion *q, fFloat angle, u_Vector *v ) {
	fFloat 			cosangle, sinangle;

	copyV( &q->v, v );
	normalizeV( &q->v );
	q->s = cos( angle/2.0 );
	scalarMV( &q->v, sin( angle/2.0 ), &q->v );
}

void zeroQ ( u_Quaternion *q ) {
	q->s = q->v.x = q->v.y = q->v.z = 0.0;
}	

void copyQ ( u_Quaternion *dest, u_Quaternion *src ) {
	memcpy( dest, src, sizeof(u_Quaternion) );
}

fFloat normQ ( u_Quaternion *q ) {
	return sqrt( q->s*q->s + q->v.x*q->v.x + q->v.y*q->v.y + q->v.z*q->v.z );
}

void normalizeQ ( u_Quaternion *q, u_Quaternion *r ) {
	fFloat 			n = normQ( q );

	r->s = q->s/n;
	r->v.x = r->v.x/n;
	r->v.y = r->v.y/n;
	r->v.z = r->v.z/n;
}

void conjugateQ ( u_Quaternion *q, u_Quaternion *r ) {
	r->s = q->s;
	r->v.x = -q->v.x;
	r->v.y = -q->v.y;
	r->v.z = -q->v.z;
}

void addQ ( u_Quaternion *q1, u_Quaternion *q2, u_Quaternion *r ) {
	r->s = q1->s + q2->s;
	r->v.x = q1->v.x + q2->v.x;
	r->v.y = q1->v.y + q2->v.y;
	r->v.z = q1->v.z + q2->v.z;
}

void subQ ( u_Quaternion *q1, u_Quaternion *q2, u_Quaternion *r ) {
	r->s = q1->s - q2->s;
	r->v.x = q1->v.x - q2->v.x;
	r->v.y = q1->v.y - q2->v.y;
	r->v.z = q1->v.z - q2->v.z;
}

void scalarMQ( u_Quaternion *q, fFloat m, u_Quaternion *r) {
	r->s = m*q->s;
	r->v.x = m*q->v.x;
	r->v.y = m*q->v.y;
	r->v.z = m*q->v.z;
}

void multQ ( u_Quaternion *q1, u_Quaternion *q2, u_Quaternion *r ) {
	u_Quaternion		t;

	t.s = q1->s*q2->s - q1->v.x*q2->v.x - q1->v.y*q2->v.y - q1->v.z*q2->v.z;
	t.v.x = q1->s*q2->v.x + q1->v.x*q2->s + q1->v.y*q2->v.z - q1->v.z*q2->v.y;
	t.v.y = q1->s*q2->v.y - q1->v.x*q2->v.z + q1->v.y*q2->s + q1->v.z*q2->v.x;
	t.v.z = q1->s*q2->v.z + q1->v.x*q2->v.y - q1->v.y*q2->v.x + q1->v.z*q2->s;
	copyQ( r, &t );
}

void inverseQ ( u_Quaternion *q1, u_Quaternion *r ) {
	conjugateQ( q1, r );
	scalarMQ( r, 1/normQ( q1 ), r );
}

/* BUG? is it correct? */
void qvq ( u_Quaternion *q1, u_Vector *v, u_Quaternion *q2, u_Vector *r ) {
	u_Quaternion 		p;

	copyV( &p.v, v );
	p.s = 0;
	multQ( q1, &p, &p );
	multQ( &p, q2, &p );
	copyV( r, &(p.v) );
}
