/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */


/* All functions here assume that vectors are allocated
   you can pass a vector simultaneously as the 'source' and as the 'result' 
   arguments in all functions 
 */

#include "mathlib.h"

u_Vector prv;

void initV ( u_Vector *vect, fFloat x, fFloat y, fFloat z ) {
	vect->x = x;
	vect->y = y;
	vect->z = z;
}

void zeroV ( u_Vector *vect ) {
	/* memset( vect, 0, sizeof(u_Vector)) is not 100% portable */
	vect->x = vect->y = vect->z = 0.0;
}

fFloat normV ( u_Vector *vect ) {
	return sqrt( vect->x*vect->x + vect->y*vect->y + vect->z*vect->z );
}

fFloat sqnormV ( u_Vector *vect ) {
	return vect->x*vect->x + vect->y*vect->y + vect->z*vect->z;
}

void copyV ( u_Vector *target, const u_Vector *src ) {
	memcpy( target, src, sizeof(u_Vector) );
}

void addV ( u_Vector *vect1, u_Vector *vect2, u_Vector *result ) {
	result->x = vect1->x + vect2->x;
	result->y = vect1->y + vect2->y;
	result->z = vect1->z + vect2->z;		
}

void subV ( u_Vector *vect1, u_Vector *vect2, u_Vector *result ) {
	result->x = vect1->x - vect2->x;
	result->y = vect1->y - vect2->y;
	result->z = vect1->z - vect2->z;		
}

void scalarMV ( u_Vector *vect, const fFloat scalar, u_Vector *result ) {
	result->x = vect->x * scalar;
	result->y = vect->y * scalar;
	result->z = vect->z * scalar;
}

void normalizeV ( u_Vector *vect ) {
	fFloat mod = normV( vect );
	if ( mod != 0 )
		scalarMV( vect, (1/mod), vect );
}

fFloat dotprodV ( u_Vector *vect1, u_Vector *vect2 ) {
	return sqrt(vect1->x*vect2->x + vect1->y*vect2->y + vect1->z*vect2->z);
}

void crossprodV ( u_Vector *vect1, u_Vector *vect2, u_Vector *result ) {
	if ( vect1 != result && vect2 != result ) {
		result->x = vect1->y*vect2->z - vect1->z*vect2->y;
		result->y = vect1->z*vect2->x - vect1->x*vect2->z;
		result->z = vect1->x*vect2->y - vect1->y*vect2->x;
	}
	else {
		prv.x = vect1->y*vect2->z - vect1->z*vect2->y;
		prv.y = vect1->z*vect2->x - vect1->x*vect2->z;
		prv.z = vect1->x*vect2->y - vect1->y*vect2->x;
		copyV( result, &prv );
	}
}

fFloat dist ( u_Vector *v1, u_Vector *v2 ) {
	subV( v2, v1, &prv );
	return( normV( &prv ) );
}

/* squared distance, faster */
fFloat sqdist ( u_Vector *v1, u_Vector *v2 ) {
	u_Vector v;
	subV( v2, v1, &v );
	return( v.x*v.x + v.y*v.y + v.z*v.z );
}
