/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file, req.c, is adapted from original code by Steve Kirkendall.
Permission given. */
#include <stdio.h>
#include <string.h>
#include "server.h"

static dFloat shorttodouble ( sShort s ) {
	return ((dFloat)s)*0.01;
#ifdef OLD
	/* integer part */
	d = ((s >> 8) & 0x7F);
	/* fractionary part, precision of 1/256 = 0,4% */
	d += ((double)(s & 0xFF))/0xFF;
	/* sign */
	d *= ( s & 0x7FFF ? -1 : 1 );
	return d;
#endif
}

/* full precision */
static dFloat longtodouble ( sLong a ) {
	return ((dFloat)a)*0.001;
#ifdef OLD
	double 		d;
	/* integer part */
	d = ( a < 0 ? -a : a );
	/* fractionary part, precision of 1/256 = 0,4% */
	d += ((double)b)/0xFFFFFFFF;
	/* sign */
	d *= ( a < 0 ? -1 : 1 );
	return d;
#endif
}

void reqhandle ( SIMPLECLIENT_UL *client, CLIENTM_UL *req ) {
	SERVERM_UL 		resp;
	SIMPLECLIENT_UL		*tempcli;
	sInt			i;
	FILE			*fp;
	UPDATE_UL		*update, *newupdate;
	OBJECT_UL		*obj;

	switch ( req->request ) {
	  case UL_HELLO:
		/* Test if we're accepting AI clients */
		if ( prefs->ai == False && req->hello.ai == True ) { /* no ai! */
			resp.response = UL_QUIT;
			netsend( client->sock, &resp );
			break;
		}
		client->driver = strdup( req->hello.driver );
		client->model = strdup( req->hello.model );
		copyColor( &client->color, &req->hello.color );

		/* send welcome */
		resp.response = UL_WELCOME;
		resp.welcome.number = (client->id>>16)&0xFFFF;
		resp.welcome.track = ptrack->name;
/**** direction, position, angle */
		netsend( client->sock, &resp );

		/* send other clients */
		for ( tempcli = firstclient; tempcli != NULL; tempcli = tempcli->next ) {
			resp.response = UL_JOINED;
			resp.joinrace.model = tempcli->model;
			resp.joinrace.driver = tempcli->driver;
			resp.joinrace.mass = tempcli->mass;
			resp.joinrace.number = (uShort)((tempcli->id >> 16) & 0xFFFF);
			copyColor( &resp.joinrace.color, &tempcli->color );
			copyV( &resp.joinrace.position, &tempcli->position );
			copyV( &resp.joinrace.direction, &tempcli->direction );
			resp.joinrace.angle = tempcli->angle;
			resp.joinrace.timestamp = timestamp( NULL );
			netsend( client->sock, &resp );

			/* and add entry to the UPDATE_UL database */
			if ( client->updates == NULL )
				update = client->updates;
			else for ( update = client->updates; update->next != NULL; 
				update = update->next ) ;
			newupdate = (UPDATE_UL *)malloc(sizeof(UPDATE_UL));
			update->next = newupdate;
			newupdate->id = client->id;
			newupdate->ts = resp.joinrace.timestamp;
			copyV( &newupdate->pos, &client->position );
			copyV( &newupdate->dir, &client->direction );
			newupdate->angle = client->angle;
			newupdate->misses = 0;
			newupdate->next = NULL;
		}
		
		/* now send objects */
		if ( general.createobject != False || 
		     general.modifyobject != False ||
		     general.destroyobject != False  ) {
			for ( obj = firstobj; obj != NULL; obj = obj->next ) {
				resp.response = UL_OBJECTCREATE;
				resp.objectcreate.name = obj->name;
				resp.objectcreate.id = obj->id;
				copyV( &resp.objectcreate.position, &obj->position );
				copyV( &resp.objectcreate.direction, &obj->direction );
				resp.objectcreate.angle = obj->angle;
				resp.objectcreate.mass = obj->mass;
				resp.objectcreate.isstatic = obj->isstatic;
				resp.objectcreate.timestamp = timestamp(NULL);
				netsend( client->sock, &resp );

				/* and add entry to the UPDATE_UL database */
				if ( client->updates == NULL )
					update = client->updates;
				else for ( update = client->updates; update->next != NULL; 
					update = update->next ) ;
				newupdate = (UPDATE_UL *)malloc(sizeof(UPDATE_UL));
				update->next = newupdate;
				newupdate->id = obj->id;
				newupdate->ts = resp.objectcreate.timestamp;
				copyV( &newupdate->pos, &client->position );
				copyV( &newupdate->dir, &client->direction );
				newupdate->angle = client->angle;
				newupdate->misses = 0;
				newupdate->next = NULL;
			}
		}

		/* update all other clients */
		resp.response = UL_JOINED;
		resp.joinrace.model = client->model;
		resp.joinrace.driver = client->driver;
		resp.joinrace.mass = client->mass;
		resp.joinrace.number = (client->id>>16)&0xFFFF;
		copyColor( &resp.joinrace.color, &client->color );
		copyV( &resp.joinrace.position, &client->position );
		copyV( &resp.joinrace.direction, &client->direction );
		resp.joinrace.angle = client->angle;
		resp.joinrace.timestamp = timestamp( NULL );
		netsend( -client->sock, &resp );
		break;

	  case UL_QUIT:
		resp.response = UL_LEFT;
		resp.leftrace.number = (client->id>>16)&0xFFFF;
		/* clean update */
		for ( tempcli = firstclient; tempcli != NULL; tempcli = tempcli->next ) {
			if ( ((((uLong)client->id)<<16)&0xFFFF0000) == tempcli->updates->id ) {
				update = client->updates;
				tempcli->updates = tempcli->updates->next;
				free( update );
			}
			else {
				for ( update = tempcli->updates; update->next->id != ((((uLong)client->id)<<16)&0xFFFF0000),
					update != NULL; update = update->next ) ;
				if ( update != NULL ) {
					newupdate = update->next;
					update->next = (newupdate == NULL ? NULL : newupdate->next );
					free( newupdate );
				}
			}
		}
		removeconnection( client->sock );
		netsend( 0, &resp );
		break;

	  case UL_CMESSAGE:
		if ( req->cmessage.to == 0 )
			i = -client->sock;
		else	i = req->cmessage.to;
		
		resp.response = UL_CMESSAGE;
		memcpy( &resp.cmessage, &req->cmessage, sizeof(req->cmessage) );
		netsend( i, &resp );
		break;

	  case UL_PAUSE:
	  case UL_UNPAUSE:
		resp.response = req->request;
		netsend( 0, &resp );
		break;

#ifdef NOTDONE
	  case UL_GETFILE:
		if ( strstr( req->control.name, TRACKEXTENSION ) ) {
			/* it's a track */
			sprintf( buffer, "%s%s%s", ULDIR, TRACKDIR,
					req->control.name );
			fp = fopen( buffer, "w" );
			/**** etc */
			fclose( fp );
		}
		netsend( client->sock, &resp );
		break;

	  case UL_FILE:
		if ( strstr( req->control.name, TRACKEXTENSION ) ) {
			sprintf( buffer, "%s%s%s", ULDIR, TRACKDIR,
					req->control.name );
			fp = fopen( buffer, "w" );
			if ( write(fp, req->control.data, req->control.size) !=
					req->control.size ) {
				fprintf( stderr, "Error: could not save %s correctly",
						req->control.name );
				/** etc */
			}
			fclose( fp );
		}
		break;
#endif

	  case UL_CONTROL:
		client->position.x += shorttodouble(req->control.dx );
		client->position.y += shorttodouble(req->control.dy );
		client->position.z += shorttodouble(req->control.dz );
		client->direction.x += shorttodouble(req->control.dwx );
		client->direction.y += shorttodouble(req->control.dwy );
		client->direction.z += shorttodouble(req->control.dwz );
		client->angle += shorttodouble(req->control.dangle );
		break;

	  case UL_PRECISECONTROL:
		copyV( &client->position, &req->precisecontrol.position );
		copyV( &client->direction, &req->precisecontrol.direction );
		client->angle = ((dFloat)req->precisecontrol.angle)*0.001;
		break;

	  case UL_COLLISION:
		resp.response = UL_COLLISION;
		memcpy( &resp.collision, &req->collision, sizeof(req->collision) );
		netsend( -client->sock, &resp );
		break;

	  case UL_OBJECTCREATE:
		obj = readObject( resp.objectcreate.id, resp.objectcreate.name,
				&resp.objectcreate.position, &resp.objectcreate.direction );
		if ( ( obj->isstatic = resp.objectcreate.isstatic ) == False ) {
			/* create new update entry */
			for ( tempcli = firstclient; tempcli != NULL; tempcli = tempcli->next ) {
				if ( client->updates == NULL )  
					update = client->updates;
				else for ( update = client->updates; update->next != NULL; 
						update = update->next ) ;

				update = (UPDATE_UL *)malloc(sizeof(UPDATE_UL));
			}
		}
		
		/* send to everyone */
		resp.response = UL_OBJECTCREATE;
		memcpy( &resp.objectcreate, &req->objectcreate, sizeof(req->objectcreate) );
		netsend( -client->sock, &resp );
		break;

	  case UL_OBJECTDESTROY:
		/* clean update */
		for ( tempcli = firstclient; tempcli != NULL; tempcli = tempcli->next ) {
			if ( resp.objectdestroy.id == client->updates->id ) {
				update = client->updates;
				client->updates = client->updates->next;
				free( update );
			}
			else { 
				for ( update = client->updates; update->next->id != resp.objectdestroy.id,
					update != NULL; update = update->next ) ;
				if ( update != NULL ) {
					newupdate = update->next;
					update->next = (newupdate == NULL ? NULL : newupdate->next );
					free( newupdate );
				}
			}
		}
		resp.response = UL_OBJECTDESTROY;
		memcpy( &resp.objectdestroy, &req->objectdestroy, sizeof(req->objectdestroy) );
		netsend( -client->sock, &resp );
		break;

	  case UL_OBJECTMODIFY:
		resp.response = UL_OBJECTMODIFY;
		memcpy( &resp.objectmodify, &req->objectmodify, sizeof(req->objectmodify) );
		netsend( -client->sock, &resp );
		break;

	  case UL_OBJECTMOVE:
		resp.response = UL_OBJECTMOVE;
		memcpy( &resp.objectmove, &req->objectmove, sizeof(req->objectmove) );
		netsend( -client->sock, &resp );
		break;

	  case UL_SPECIAL:
		if ( req->special.to == 0 )
			i = -client->sock;
		else	i = req->cmessage.to;
		
		resp.response = UL_SPECIAL;
		memcpy( &resp.special, &req->special, sizeof(req->special) );
		netsend( i, &resp );
		break;

	  case UL_FOO:
		/* nothing */
		break;
		
	  default:
		fprintf( stderr, "handlereq: unknown request %d\n",
			  req->request);
	}
}
