/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file, resp.c, is adapted from original code by Steve Kirkendall.
Permission given. */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "server.h"

static sInt 	lsock;			/* new connections */
sShort		nclients, actual;	/* nclients is the total number of 
					   clients in the sim, while actual is
					   the number of the next client to
					   enter the simulation */
TRACK_UL 	*ptrack;
STATUS_UL	servers;
SIMPLECLIENT_UL	*firstclient;
GAMEMODE_UL	mode;


/*** we should read trackname here */
static sInt handleArgs ( sInt argc, sByte **argv ) {
	sInt			i;
	uByte			buffer[200], *t;
	
	ptrack = NULL;

	if ( argc == 1 ) {
		fflush( stderr );
		fflush( stdout );
		printf( "\nChoose port [%d]: ", DEFPORT );
		fgets( buffer, 200, stdin );
		fflush( stdin );
		t = buffer;
		while ( isspace(*t) ) t++;
		if ( isdigit(*t) )
			prefs->port = atoi( buffer );
		
		printf( "\nChoose game mode [FOREVER/match]: " );
		fgets( buffer, 200, stdin );
		fflush( stdin );
		if ( strstr( "match", buffer ) || strstr( "MATCH", buffer ) )
			mode = MATCH;
		else
			mode = FOREVER;

		do {
			printf( "\nChoose track [press enter for list]: " );
			fgets( buffer, 200, stdin );
			fflush( stdin );
			if ( buffer[0] == '\n' ) {
				sprintf( buffer, "ls %s%s | more", ULDIR, TRACKDIR );
				if ( system (NULL) )
					system( buffer );
				else
					printf( "Could not list files.\n" );
			}
			else {
				t = buffer;
				while ( *t != '\n' ) t++;
				t = '\0';
				t = buffer;
				while ( isspace(*t) ) t++;
				ptrack = readTrack( t );
			}
		} while ( ptrack == NULL );

		printf( "\n" );
	}
	else {
		for ( i = 1; i < argc; i++ ) {
			if ( !strcmp( argv[i], "-port" ) ) {
				prefs->port = atoi( argv[++i] );
				fprintf( stdout, "Using port %d\n", prefs->port );
			}
			else if ( !strcmp( argv[i], "-m" ) ) {
				if ( !strcmp( argv[i], "forever" ) )
					mode = FOREVER;
				else if ( !strcmp( argv[i], "match" ) )
					mode = MATCH;
				else
					goto error;
			}
			else {
error:
				fprintf( stdout, "%s%d%s",
					"\nArguments are:\n"
					"  -port number\t\tport to be used\n"
					"  -mode [forever|match]\tgame mode\n"
					"  -track name\t\ttrack to be used\n"
					"If no arguments are given, a interactive setup is used.\n" );
				return FATALERROR;
			}
		}
	}
	return OK;
}

/* add a new connection */
static void addconnection ( sInt sock ) {
	SIMPLECLIENT_UL		*new, *tmp;
	Bool			b;

#ifdef DEBUG
	fprintf(stderr, "addconnection(%d)\n", sock);
#endif

	/* create a connection to an unknown user */
	new = (SIMPLECLIENT_UL *)malloc(sizeof(SIMPLECLIENT_UL));
	new->sock = sock;
	if ( !firstclient )
		firstclient = new;
	else {
		for ( tmp = firstclient; tmp->next != NULL; tmp = tmp->next ) ;
		tmp->next = new;
		new->next = (SIMPLECLIENT_UL *)NULL;
	}
	
	/* set new id, checking for its uniqueness */
	new->id = 0;
	actual++;
	b = False;
	do {
		clientloop:
		for ( tmp = firstclient; tmp->next != NULL; tmp = tmp->next ) {
			if ( (((uLong)actual)<<16)&0xFFFF0000 == tmp->id ) {
				actual++;
				goto clientloop;
			}
		}
		b = True;
	} while ( b == False );
	new->id = (((uLong)actual)<<16)&0xFFFF0000;

	nclients++;
	new->status = READY;
	/* add the socket to the list checked by nrselect() */
	netaddsock( sock );

	/* UPDATE_UL structure is done in req.c:reqhandle() */
}

void removeconnection ( sInt sock ) {
	SIMPLECLIENT_UL		*oldc, *tmp;
	UPDATE_UL		*update, *update2;
	uLong			id;

	/* find the connection */
	for ( tmp = NULL, oldc = firstclient;
	     oldc->sock != sock;
	     tmp = oldc, oldc = oldc->next) ;

	/* save id for later use */
	id = oldc->id;

	/* reset it -- free whatever it allocated, etc. */
	freeDoodad( oldc->geom );
	free( oldc->model );
	free( oldc->driver );
	/* free update information */
	for ( update = oldc->updates; update != NULL; ) {
		update2 = update->next;
		free( update );
		update = update2; 
	}

	/* close its socket */
	netclose( sock );
	netdelsock( sock );

	/* remove it from the list of clients */
	if ( tmp )
		tmp->next = oldc->next;
	else
		firstclient = oldc->next;
	nclients--;

	/* free its memory */
	free( oldc );

	/* free it's update entry in other clients */
	for ( tmp = firstclient; tmp != NULL; tmp = tmp->next ) {
		for ( update = NULL, update2 = tmp->updates; 
				update2->id != id, update2 != NULL; ) {
			update = update2;
			update2 = update2->next;
		}
		update->next = update2->next;
		free( update2 );
	}
}

/* send updates. Missing precisecontrol/objectmodify. Some extra parameter
fields should be added to SIMPLECLIENT_UL too, such as when to send precise
positions, etc. Finally, this thing can be optimized; it's actually making
all calculations even if they aren't needed. */
static void sendupdates ( void ) {
	SIMPLECLIENT_UL 	*client, *tmpc;
	OBJECT_UL		*obj;
	TIMESTAMP_UL		timenow;
	SERVERM_UL		resp;
	UPDATE_UL		*update, *up2;
	VECTOR_UL		v;
	
	for ( client = firstclient; client != NULL; client = client->next, 
			timenow = timestamp( NULL ) ) {
	   for ( update = client->updates; update != NULL; update = update->next ) {
		subV( &client->position, &update->pos, &v );
		/* check if inside distance range & if it's time to send
		 update */
		if ( client->maxsqdist > dotprodV( &v, &v ) &&
			(update->ts - timenow) > client->secsperupdate ) {
			if ( update->id & 0xFFFF ) { /*object */
				resp.response = UL_OBJECTMOVE;
				for ( obj = firstobj; obj->id != update->id, obj != NULL;
						 obj = obj->next ) ;
				if ( obj == NULL ) {
					fprintf( stderr, "Object not found!\n" );
					for ( up2 = client->updates; up2->next != update; up2 = up2->next ) ;
					up2->next = update->next;
					free( update );
				}

				resp.objectmove.id = update->id;
				subV( &obj->position, &update->pos, &v );
				resp.objectmove.dx = (sShort)v.x*100;
				resp.objectmove.dy = (sShort)v.y*100;
				resp.objectmove.dz = (sShort)v.z*100;
				copyV( &update->pos, &obj->position );
				subV( &obj->direction, &update->dir, &v );
				resp.objectmove.dwx = (sShort)v.x*100;
				resp.objectmove.dwy = (sShort)v.y*100;
				resp.objectmove.dwz = (sShort)v.z*100;
				copyV( &update->pos, &obj->position );
				resp.objectmove.dangle = (sShort)(obj->angle-update->angle)*100;
				update->angle = obj->angle;
				if ( resp.objectmove.dx == 0 && resp.objectmove.dy == 0 &&
					resp.objectmove.dz == 0 && resp.objectmove.dwx == 0 &&
					resp.objectmove.dwy == 0 && resp.objectmove.dwz == 0 &&
					resp.objectmove.dangle == 0 )
					continue;

				resp.objectmove.timestamp = (uShort)(timenow & 0xFFFF);
			}
			else { /* client */
				resp.response = UL_CONTROL;
				for ( tmpc = firstclient; tmpc->id != update->id, tmpc != NULL;
						tmpc = tmpc->next ) ;
				if ( tmpc == NULL ) {
					fprintf( stderr, "Client not found!\n" );
					for ( up2 = client->updates; up2->next != update; up2 = up2->next ) ;
					up2->next = update->next;
					free( update );
				}

				resp.control.number = update->id >> 16;
				subV( &tmpc->position, &update->pos, &v );
				resp.control.dx = (sShort)v.x*100;
				resp.control.dy = (sShort)v.y*100;
				resp.control.dz = (sShort)v.z*100;
				copyV( &update->pos, &tmpc->position );
				subV( &client->direction, &update->dir, &v );
				resp.control.dwx = (sShort)v.x*100;
				resp.control.dwy = (sShort)v.y*100;
				resp.control.dwz = (sShort)v.z*100;
				copyV( &update->dir, &tmpc->direction );
				resp.control.dangle = (sShort)(tmpc->angle-update->angle)*100;
				update->angle = tmpc->angle;
				if ( resp.control.dx == 0 && resp.control.dy == 0 &&
					resp.control.dz == 0 && resp.control.dwx == 0 &&
					resp.control.dwy == 0 && resp.control.dwz == 0 &&
					resp.control.dangle == 0 ) 
					continue;

				resp.control.timestamp = (uShort)(timenow & 0xFFFF);
			}
			netsend( client->sock, &resp );
		}
	    }
	}
}

sInt main ( sInt argc, sByte **argv ) {
	SIMPLECLIENT_UL		*client;
	sInt			port, sock;
	sLong			timeout;
	CLIENTM_UL		*req;
	SERVERM_UL		resp;

	actual = nclients = 0;

	fprintf( stdout, "UNLIMITED SIMULATOR SERVER, Version %s %s\n",  
			VERSION, BUILDDATE );

	if ( loadprefs() == FATALERROR )
		exit(1);

	if ( handleArgs( argc, argv ) == FATALERROR ) 
		exit(2);

	inittime();
	
	lsock = netlisten( port );
	if ( lsock < 0 )
		exit(3);
	
	netaddsock( lsock );
	
	sock = netwait(-1L);
	if ( sock != lsock || (sock = netaccept(lsock)) < 0 ) {
		fprintf( stderr, "Could not connect to first client\n" );
		exit(4);
	}
	addconnection( sock );

	do {
		do {
			if ( mode == MATCH ) {
				switch ( servers ) {
				   case CONNECTING:
					timeout = -1;
					break;
				   case COUNTDOWN:
					timeout = 1000;
					break;
				   default:
					timeout = 50;
				}
			}
			netflush();

			sock = netwait(timeout);
			if ( sock == lsock ) {
				sock = netaccept(lsock);
				if ( sock >= 0 && 
					((mode == MATCH && servers == CONNECTING) 
					|| mode == FOREVER) )
						addconnection(sock);
				else if ( sock >= 0 ) 
					/* make sure it's skipped. Future:
					send a message notifying that we're not
					accepting connections */
					goto skip;
			}
			else if ( sock >= 0 ) {
				/* must be a client request (or requests) */
				/* find the client */
				for ( client = firstclient;
				     client->sock != sock, client != NULL;
				     client = client->next ) ;

				if ( client != NULL ) {
					/* handle its requests */
					do {
						req = netrcv(sock);
						if (!req)
							removeconnection(sock);
						else
							reqhandle( client, req );
					} while ( netmore() );
				}
				else {
					fprintf( stderr, 
						"Message received from unknown client, ignoring\n" );
				}
			}
skip:
			sendupdates();

		} while ( sock >= 0 && firstclient != NULL );

		/* If we get here, either there was an error or we timed out.
		 * Assume the latter.  Time to update every car!
		 */
		if ( firstclient && servers == PAUSED ) {
			netflush();
		}

/********* gotta make this below work, following the guidelines of
		general.lyx */
		else if ( firstclient && servers == COUNTDOWN ) {
#ifdef NTOTKW
			/* tell everybody that 1 second has elapsed */
			resp.response = COUNTDOWN;
			resp.countdown.time = 1000;
			netsend( 0, &resp );
			netflush();
			if ( countdown == 0 )
				servers = RACING;
#endif
		}
	} while ( (mode == FOREVER ? (0 == 0) : nclients) );

	return 0;
}
